#!/usr/bin/env python

import os, subprocess
from tempfile import mkstemp
try:
    from opaquestore import client
    available = True
except:
    client = None
    available = False

if available:
    client.config = client.processcfg(client.getcfg('opaque-store'))

def usage(params):
  print("\nOPAQUE Store style blobs")
  print("       echo -n 'password' | %s store <keyid> file-to-store" % params[0])
  print("       echo -n 'password' | %s read <keyid>" % params[0])
  print("       echo -n 'password' | %s replace [force] <keyid> file-to-store" % params[0])
  print("       echo -n 'password' | %s edit [force] <keyid>" % params[0])
  print("       echo -n 'password' | %s changepwd [force] <keyid>" % params[0])
  print("       echo -n 'password' | %s erase [force] <keyid>" % params[0])
  print("       echo -n 'password' | %s recovery-tokens <keyid>" % params[0])
  print("       echo -n 'password' | %s unlock <keyid> <recovery-token>" % params[0])

def is_cmd(params):
    if params[1] not in cmds: return False
    return True

def connect():
  s = client.Multiplexer(client.config['servers'])
  s.connect()
  return s

def store(pwd, keyid, path):
  # create & recovery-tokens
  with open(path,'r') as fd:
    data = fd.read()
  s = connect()
  client.create(s, pwd, keyid.encode('utf8'), data.encode('utf8'))
  s = connect()
  token = client.get_recovery_tokens(s, pwd, keyid.encode('utf8'))
  print("successfully created opaque store record. Store the following recovery token, in case this record is locked")
  print(token)

def read(pwd, keyid):
  s = connect()
  print(client.get(s, pwd, keyid.encode('utf8')))

def replace(pwd, keyid, path, force=False):
  with open(path,'r') as fd:
    data = fd.read()
  s = connect()
  client.update(s, pwd, keyid.encode('utf8'), data.encode('utf8'), force)

def erase(pwd, keyid, ctx, force=False):
  s = connect()
  client.delete(s, pwd, keyid.encode('utf8'), force)
  # also handle delete sphinx record in case ostore.erase
  m = ctx['m'](ctx['servers'])
  m.connect()
  ctx['delete'](m, ctx['pwd'], ctx['user'], ctx['host'])

def changepwd(pwd, keyid, ctx, force=False):
  m = ctx['m'](ctx['servers'])
  m.connect()
  newpwd = ctx['change'](m, ctx['pwd'], ctx['newpwd'], ctx['user'], ctx['host'])

  s = connect()
  data = client.get(s, pwd, keyid.encode('utf8'))

  s = connect()
  client.delete(s, pwd, keyid.encode('utf8'), force)

  s = connect()
  client.create(s, newpwd, keyid.encode('utf8'), data.encode('utf8'))

  s = connect()
  token = client.get_recovery_tokens(s, newpwd, keyid.encode('utf8'))

  m = ctx['m'](ctx['servers'])
  m.connect()
  ctx['commit'](m, ctx['pwd'], ctx['user'], ctx['host'])
  print("Store the following recovery token, in case this record is locked")
  print(token)

def edit(pwd, keyid, force=False):
  if not os.path.exists('/dev/tty'):
      print("can only edit on systems that have /dev/tty, sorry")
      return False
  s = connect()
  data = client.get(s, pwd, keyid.encode('utf8'))
  fd, fname = mkstemp()
  fd = os.fdopen(fd,'w')
  fd.write(data)
  fd.close()
  tty = os.open("/dev/tty", os.O_RDWR|os.O_LARGEFILE)
  subprocess.run([os.environ.get("EDITOR"), fname], stdin=tty, stdout=tty, stderr=tty)
  with open(fname,"r") as fd:
    data = fd.read()
  os.unlink(fname)
  s = connect()
  client.update(s, pwd, keyid.encode('utf8'), data.encode('utf8'), force)

def recoverytoken(pwd, keyid):
  s = connect()
  token = client.get_recovery_tokens(s, pwd, keyid.encode('utf8'))
  print("Store the following recovery token, in case this record is locked")
  print(token)

def unlock(pwd, keyid, token):
  # unlock + get
  s = connect()
  client.unlock(s, token, keyid.encode('utf8'))
  s = connect()
  print(client.get(s, pwd, keyid.encode('utf8')))

cmds = {'store': store,
        'read': read,
        'replace': replace,
        'edit': edit,
        'changepwd': changepwd,
        'erase': erase,
        'recovery-tokens': recoverytoken,
        'unlock': unlock}

def parse(params):
    if params[1] not in cmds: return False

    op = cmds[params[1]]
    args = []

    if params[1] in {'replace', 'edit', 'erase', 'changepwd'} and params[2]=='force':
        del params[2]
        args.append(True)

    keyid=params[2]

    if params[1] in {'store', 'replace'}:
        if not os.path.isfile(params[3]):
            raise ValueError(f'opaque store parameter "{params[3]}" is not a file, how would i store it?')
        args.insert(0, params[3])

    if params[1] == 'unlock':
        args.insert(0,params[3])

    return op, keyid, args



#!/usr/bin/env python

from pwdsphinx import bin2pass
import traceback, os, sys
from importlib.machinery import SourceFileLoader
from pathlib import Path

converters = {}

def load_converters():
  global converters
  p = Path(__file__).parent.absolute()
  for converter_fname in os.listdir(f'{p}/converters/'):
    if converter_fname.startswith('_') or not converter_fname.endswith('.py'):
        continue
    try:
      name = converter_fname[:-3]
      import_path = 'converters.'+name
      if import_path in sys.modules:
          del sys.modules[import_path]
      s = SourceFileLoader(import_path,
                           f'{p}/converters/' + converter_fname).load_module()
    except:
      print("failed to load converter", converter_fname)
      traceback.print_exc()
      continue
    for schema, converter in s.schema.items():
      if schema in converters:
        raise ValueError(f"{schema} is a already in loaded converters")
      converters[schema]=converter

load_converters()

def convert(rwd, user, *opts):

  if '://' not in user:
    return bin2pass.derive(rwd, *opts)
  elif user.startswith('otp://'):
    # need to recover the predefined base32 otp key
    rwd = bin2pass.derive(rwd, *opts)

  schema, _ = user.split("://",1)
  return converters[schema](rwd, *opts)

if __name__ == "__main__":
  convert(b'\xaa' * 32, 'asdf', 'uld', 0, '')
  import sys
  args = [int(x) if x.isdigit() else x for x in sys.argv[1:]]
  rwd = sys.stdin.buffer.read(32)
  print(convert(rwd, *args))



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2018-2024, Marsiske Stefan
# SPDX-License-Identifier: GPL-3.0-or-later

import sys, os, socket, ssl, struct, time
from SecureString import clearmem
from equihash import solve
import pysodium
try:
  from pwdsphinx import bin2pass
  from pwdsphinx.config import getcfg
  from pwdsphinx.consts import *
  from pwdsphinx.converter import convert
except ImportError:
  import bin2pass
  from config import getcfg
  from consts import *
  from converter import convert

# override consts from consts.py
VERSION = b'\x00'
RULE_SIZE = 79

#### config ####

cfg = getcfg('sphinx')
enabled=False
verbose = cfg['client'].get('verbose', False)
hostname = cfg['client'].get('address')
if hostname is not None:
   enabled = True
   address = socket.gethostbyname(hostname)
   port = int(cfg['client'].get('port',2355))
   try:
     ssl_cert = os.path.expanduser(cfg['client'].get('ssl_cert')) # only for dev, production system should use proper certs!
   except TypeError: # ignore exception in case ssl_cert is not set, thus None is attempted to expand.
     ssl_cert = None

datadir = os.path.expanduser(cfg['client'].get('datadir','~/.config/sphinx'))
#  make RWD optional in (sign|seal)key, if it is b'' then this protects against
#  offline master pwd bruteforce attacks, drawback that for known (host,username) tuples
#  the seeds/blobs can be controlled by an attacker if the masterkey is known
rwd_keys = cfg['client'].get('rwd_keys', False)
validate_password = cfg['client'].get('validate_password',True)
userlist = cfg['client'].get('userlist', True)

if verbose and enabled:
   print("v1 hostname:", hostname, file=sys.stderr)
   print("v1 address:", address, file=sys.stderr)
   print("v1 port:", port, file=sys.stderr)
   print("v1 ssl_cert:", ssl_cert, file=sys.stderr)

#### consts ####

ENC_CTX = b"sphinx encryption key"
SIGN_CTX = b"sphinx signing key"
SALT_CTX = b"sphinx host salt"
PASS_CTX = b"sphinx password context"
CHECK_CTX = b"sphinx check digit context"

#### Helper fns ####

def get_masterkey():
  try:
    with open(os.path.join(datadir,'masterkey'), 'rb') as fd:
        mk = fd.read()
    return mk
  except FileNotFoundError:
    raise ValueError("ERROR: Could not find masterkey!\nIf sphinx was working previously it is now broken.\nIf this is a fresh install all is good, you just need to run `%s init`." % sys.argv[0])

def connect():
  ctx = ssl.create_default_context()
  if(ssl_cert):
      ctx.load_verify_locations(ssl_cert) # only for dev, production system should use proper certs!
      ctx.check_hostname=False            # only for dev, production system should use proper certs!
      ctx.verify_mode=ssl.CERT_NONE       # only for dev, production system should use proper certs!
  else:
      ctx.load_default_certs()
      ctx.verify_mode = ssl.CERT_REQUIRED
      ctx.check_hostname = True

  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  s.settimeout(5)
  s = ctx.wrap_socket(s, server_hostname=hostname)
  s.connect((address, port))
  return s

def get_signkey(id, rwd):
  mk = get_masterkey()
  seed = pysodium.crypto_generichash(SIGN_CTX, mk)
  clearmem(mk)
  # rehash with rwd so the user always contributes his pwd and the sphinx server it's seed
  seed = pysodium.crypto_generichash(seed, id)
  if rwd_keys:
    seed = pysodium.crypto_generichash(seed, rwd)
  pk, sk = pysodium.crypto_sign_seed_keypair(seed)
  clearmem(seed)
  return sk, pk

def get_sealkey():
  mk = get_masterkey()
  sk = pysodium.crypto_generichash(ENC_CTX, mk)
  clearmem(mk)
  return sk

def encrypt_blob(blob):
  # todo implement padding to hide length information
  sk = get_sealkey()
  nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
  ct = pysodium.crypto_aead_xchacha20poly1305_ietf_encrypt(blob,VERSION,nonce,sk)
  clearmem(sk)
  return VERSION+nonce+ct

def decrypt_blob(blob):
  # todo implement padding to hide length information
  sk = get_sealkey()
  version = blob[:1]
  if version > VERSION:
    raise ValueError("Your client is too old to handle this response. Please update your client.")
  blob = blob[1:]
  nonce = blob[:pysodium.crypto_secretbox_NONCEBYTES]
  blob = blob[pysodium.crypto_secretbox_NONCEBYTES:]
  res = pysodium.crypto_aead_xchacha20poly1305_ietf_decrypt(blob,version,nonce,sk)
  clearmem(sk)
  return version, res

def sign_blob(blob, id, rwd):
  sk, pk = get_signkey(id, rwd)
  res = pysodium.crypto_sign_detached(blob,sk)
  clearmem(sk)
  return b''.join((blob,res))

def getid(host, user):
  mk = get_masterkey()
  salt = pysodium.crypto_generichash(SALT_CTX, mk)
  clearmem(mk)
  return pysodium.crypto_generichash(b'|'.join((user.encode(),host.encode())), salt, 32)

def unpack_rule(ct):
  version, packed = decrypt_blob(ct)
  xor_mask = packed[-32:]
  v = int.from_bytes(packed[:-32], "big")

  size = v & ((1<<7) - 1)
  rule = {c for i,c in enumerate(('u','l','d')) if (v >> 7) & (1 << i)}
  symbols = [c for i,c in enumerate(bin2pass.symbols) if (v>>(7+3) & (1<<i))]
  if validate_password:
      check_digit = (v>>(7+3+33))
  else:
      check_digit = 0

  return rule, symbols, size, check_digit, xor_mask

def xor(x,y):
  return bytes(a ^ b for (a, b) in zip(x, y))

def ratelimit(s,req):
  pkt0 = b''.join([CHALLENGE_CREATE, req])
  s.send(pkt0)
  challenge = s.recv(1+1+8+32) # n,k,ts,sig
  if len(challenge)!= 1+1+8+32:
    if verbose: print("challengelen incorrect: %s %s" %(len(challenge), repr(challenge)), file=sys.stderr)
    raise ValueError("ERROR: failed to get ratelimit challenge")
  s.close()
  n = challenge[0]
  k = challenge[1]

  try:
    os.write(3,f"{n} {k}\n".encode('utf8'))
  except OSError: pass

  if k==4:
    if n < 90:
      if verbose: print("got an easy puzzle: %d" % n, file=sys.stderr)
    elif n > 100:
      if verbose: print("got a hard puzzle: %d" % n, file=sys.stderr)
    else:
      if verbose: print("got a moderate puzzle: %d" % n, file=sys.stderr)
  seed = challenge + req

  delta = time.time()
  solution = solve(n, k, seed)
  delta = time.time() - delta
  try:
    os.write(3,f"{delta}".encode('utf8'))
  except OSError: pass

  s = connect()
  pkt1 = b''.join([CHALLENGE_VERIFY, challenge])
  s.send(pkt1)
  s.send(req)
  s.send(solution)
  return s

def auth(s,id,alpha=None,pwd=None,r=None):
  if r is None:
    nonce = s.recv(32)
    if len(nonce)!=32:
       return False
    rwd = b''
  else:
    msg = s.recv(64)
    if len(msg)!=64:
       return False
    beta = msg[:32]
    nonce = msg[32:]
    rwd = finish(pwd, r, alpha, beta, id)

  sk, pk = get_signkey(id, rwd)
  clearmem(rwd)
  sig = pysodium.crypto_sign_detached(nonce,sk)
  clearmem(sk)
  s.send(sig)

  resp = s.recv(6)
  if resp==b'\x00\x04auth': return True
  return False

def challenge(pwd):
  h0 = pysodium.crypto_generichash(pwd, outlen=pysodium.crypto_core_ristretto255_HASHBYTES);
  H0 = pysodium.crypto_core_ristretto255_from_hash(h0)
  clearmem(h0)
  r = pysodium.crypto_core_ristretto255_scalar_random()
  alpha = pysodium.crypto_scalarmult_ristretto255(r, H0)
  clearmem(H0)
  return r, alpha

def finish(pwd, r, alpha, beta, salt):
  if(alpha==beta): raise ValueError("alpha == beta")
  if not pysodium.crypto_core_ristretto255_is_valid_point(alpha): raise ValueError("invalid beta")

  rinv = pysodium.crypto_core_ristretto255_scalar_invert(r)
  H0_k = pysodium.crypto_scalarmult_ristretto255(rinv, beta)
  clearmem(rinv)
  rwd0 = pysodium.crypto_generichash(pwd+H0_k, outlen=pysodium.crypto_core_ristretto255_BYTES);
  clearmem(H0_k)
  rwd = pysodium.crypto_pwhash(pysodium.crypto_core_ristretto255_BYTES,
                               rwd0, salt[:pysodium.crypto_pwhash_SALTBYTES],
                               pysodium.crypto_pwhash_OPSLIMIT_INTERACTIVE,
                               pysodium.crypto_pwhash_MEMLIMIT_INTERACTIVE)
  clearmem(rwd0)
  return rwd

#### OPs ####

def get(pwd, user, host):
  if isinstance(pwd, str): pwd = pwd.encode()
  s = connect()
  id = getid(host, user)
  r, alpha = challenge(pwd)
  msg = b''.join([V1GET, id, alpha])
  s = ratelimit(s, msg)

  resp = s.recv(32+RULE_SIZE) # beta + sealed rules
  if resp == b'\x00\x04fail' or len(resp)!=32+RULE_SIZE:
      s.close()
      raise ValueError("ERROR: Either the record does not exist, or the request to server was corrupted during transport.")
  beta = resp[:32]
  rules = resp[32:]
  rwd = finish(pwd, r, alpha, beta, id)

  try:
    classes, symbols, size, checkdigit, xormask = unpack_rule(rules)
  except ValueError:
    s.close()
    raise ValueError("ERROR: failed to unpack password rules from server")
  s.close()

  if validate_password and (checkdigit != (pysodium.crypto_generichash(CHECK_CTX, rwd, 1)[0] & ((1<<5)-1))):
    raise ValueError("ERROR: bad checkdigit")

  rwd = xor(pysodium.crypto_generichash(PASS_CTX, rwd),xormask)
  ret = convert(rwd,user,classes,size,symbols)
  clearmem(rwd)

  return ret

def delete(pwd, user, host):
  s = connect()
  # run sphinx to recover rwd for authentication
  id = getid(host, user)
  r, alpha = challenge(pwd)
  msg = b''.join([V1DELETE, id, alpha])
  s = ratelimit(s, msg)

  if isinstance(pwd, str): pwd = pwd.encode()
  if not auth(s,id,alpha,pwd,r):
    s.close()
    raise ValueError("ERROR: Failed to authenticate to server while deleting password on server or record doesn't exist")

  if not userlist:
     s.send(b"\0"*96)
  else:
    # delete user from user list for this host
    # a malicous server could correlate all accounts on this services to this users here
    # first query user record for this host
    id = getid(host, '')
    signed_id = sign_blob(id, id, b'')
    s.send(signed_id)
    # wait for user blob
    bsize = s.recv(2)
    bsize = struct.unpack('!H', bsize)[0]
    if bsize == 0:
      # this should not happen, it means something is corrupt
      s.close()
      raise ValueError("ERROR: server has no associated user record for this host", file=sys.stderr)

    blob = s.recv(bsize)
    if blob == b'fail':
      s.close()
      raise ValueError("ERROR: invalid signature on list of users")
    version, blob = decrypt_blob(blob)
    users = set(blob.decode().split('\x00'))
    if user not in users:
      # this should not happen, but maybe it's a sign of corruption?
      s.close()
      raise ValueError(f'warning "{user}" is not in user record', file=sys.stderr)
    users.remove(user)
    blob = ('\x00'.join(sorted(users))).encode()
    # notice we do not add rwd to encryption of user blobs
    blob = encrypt_blob(blob)
    bsize = len(blob)
    if bsize >= 2**16:
      s.close()
      raise ValueError("ERROR: blob is bigger than 64KB.")
    blob = struct.pack("!H", bsize) + blob
    blob = sign_blob(blob, id, b'')

    s.send(blob)

  if b'ok' != s.recv(2):
    s.close()
    raise ValueError("ERROR: server failed to save updated list of user names for host: %s." % host)

  s.close()
  return True

def read_blob(s, id):
  msg = b''.join([READ, id])
  s = ratelimit(s, msg)
  if auth(s,id) is False:
    s.close()
    return
  bsize = s.recv(2)
  bsize = struct.unpack('!H', bsize)[0]
  blob = s.recv(bsize)
  s.close()
  if blob == b'fail':
    return
  return decrypt_blob(blob)

def users(host):
  s = connect()
  res = read_blob(s, getid(host, ''))
  if not res: return set()
  version, res = res
  users = set(res.decode().split('\x00'))
  return users

#print(get(connect(), b'asdf','asdf','test'))
#print(delete(connect(), b'asdf','asdf','test'))



#!/usr/bin/env python3

def convert(rwd, *opts):
    size = opts[1]
    # rwd[:] does not copy the underlying data, and thus
    # a clearmem() not only wipes the original, but also the copy...
    return rwd[:1] + rwd[1:size]

schema = {"raw": convert}

def main():
    for rwd in sys.stdin:
        print(convert(rwd.strip()))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

from base64 import b32decode
import hmac
from struct import pack, unpack
import sys
from time import time

def totp(key, *opts, time_step=30, digits=6, digest='sha1'):
    if isinstance(key, bytes): key=key.decode('utf8')
    key = b32decode(key.upper() + '=' * ((8 - len(key)) % 8))
    ts = pack('>Q', int(time() / time_step))
    mac = hmac.new(key, ts, digest).digest()
    offset = mac[-1] & 0x0f
    binary = unpack('>L', mac[offset:offset+4])[0] & 0x7fffffff
    return str(binary)[-digits:].zfill(digits)

schema = {"otp": totp}

def main():
    args = [int(x) if x.isdigit() else x for x in sys.argv[1:]]
    for key in sys.stdin:
        print(totp(key.strip(), [], *args))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import unittest, random, math
from pwdsphinx import bin2pass

# to get coverage, run
# PYTHONPATH=.. coverage run ../tests/pass2bin.py
# coverage report -m
# to just run the tests do
# python3 -m unittest discover --start-directory ../tests

class TestRules(unittest.TestCase):
    def test_invert_simple(self):
      target = "this is a pass2bin test string"
      rwd, classes, symbols = bin2pass.pass2bin(target)
      self.assertEqual(bin2pass.derive(rwd, classes, len(target), symbols), target)

    def test_invert_too_long(self):
      target = "this is a pass2bin test stringthis is a pass2bin test stringthis is a pass2bin test stringthis is a pass2bin test stringthis is a pass2bin test string"
      self.assertRaises(OverflowError, bin2pass.pass2bin, target)

    def test_invert_iter(self):
      chars = bin2pass.allchars
      for i in range(1,len(chars)):
        target=''.join(chars[:i])
        try:
          rwd, classes, symbols = bin2pass.pass2bin(target)
        except OverflowError:
          break
        self.assertEqual(bin2pass.derive(rwd, classes, len(target), symbols), target)

    def test_invert_reviter(self):
      chars = bin2pass.allchars
      for i in range(1,len(chars)):
        target=''.join(chars[-i:])
        try:
          rwd,classes, symbols = bin2pass.pass2bin(target)
        except OverflowError:
          break
        self.assertEqual(bin2pass.derive(rwd, classes, len(target), symbols), target)

    def test_invert_random(self):
      chars = bin2pass.allchars
      for _ in range(1000):
         target=''.join(random.choices(chars,k=random.randrange(1,39)))
         rwd,classes, symbols = bin2pass.pass2bin(target)
         self.assertEqual(bin2pass.derive(rwd, classes, len(target), symbols), target)

    def test_all_zeroes(self):
        logbase = int(math.log(1<<256, len(bin2pass.allchars)))
        target = bin2pass.allchars[0] * (logbase-1) + bin2pass.allchars[1]
        for _ in range(logbase):
          rwd,classes, symbols = bin2pass.pass2bin(target)
          self.assertEqual(bin2pass.derive(rwd, classes, len(target), symbols), target)
          target = target[1:]+bin2pass.allchars[0]

    def test_short_zeroes(self):
        logbase = int(math.log(1<<256, len(bin2pass.allchars)))
        target = bin2pass.allchars[0] * (logbase//2) + bin2pass.allchars[1]
        for _ in range(len(target)):
          ctarget = ''.join(target)
          (rwd,classes,symbols) = bin2pass.pass2bin(ctarget)
          self.assertEqual(bin2pass.derive(rwd, classes, len(ctarget), symbols), ctarget)
          target = target[1:]+bin2pass.allchars[0]

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import unittest
from pwdsphinx.converter import convert
from pwdsphinx import bin2pass

char_classes = 'uld'
symbols = bin2pass.symbols
size = 0

class TestConverters(unittest.TestCase):
  def test_bin2pass(self):
    rwd = b'\xaa' * 32
    pwd = convert(rwd, "asdf", char_classes, size, symbols)
    self.assertEqual(pwd, '2UH@/%XoTb+T-RT*tipUqT+b\'lYQ*kUiPOdq@sK')

  def test_bin2pass_8char(self):
    rwd = b'\xaa' * 32
    pwd = convert(rwd, "asdf", char_classes, 8, symbols)
    self.assertEqual(pwd, 'iPOdq@sK')

  def test_raw(self):
    rwd = b'\xaa' * 32
    pwd = convert(rwd, "raw://asdf", char_classes, len(rwd), symbols)
    self.assertEqual(pwd, rwd)


  def test_otp(self):
    pwd = 'A' * 16
    rwd, classes, symbols = bin2pass.pass2bin(pwd)
    pwd = convert(rwd, "otp://asdf", classes, len(pwd), symbols)
    self.assertIsInstance(pwd, str)
    self.assertEqual(len(pwd), 6)

  def test_age(self):
    rwd = b'\x55' * 32
    pwd = convert(rwd, "age://asdf", char_classes, size, symbols)
    self.assertEqual(pwd, 'AGE-SECRET-KEY-1242424242424242424242424242424242424242424242424242S6JN5PD')
    from pwdsphinx.converters.sphage import decode
    self.assertEqual(rwd, bytes(decode('age-secret-key-', pwd)))



from os import mkdir, getenv
from os.path import exists, split
from pathlib import Path
from sys import executable

EXT_NM_TPL = '''{{
  "name": "websphinx",
  "description": "Host for communicating with PITCHFORKed Sphinx",
  "path": "{cmd}",
  "type": "stdio",
'''

CHR_NM_TPL = EXT_NM_TPL + '''  "allowed_origins": [
    "chrome-extension://ojbhlhidchjkmjmpeonendekpoacahni/"
  ]
}}
'''

FF_NM_TPL = EXT_NM_TPL + '''  "allowed_extensions": [
    "sphinx@ctrlc.hu"
  ]
}}
'''


def get_executable():
    venv = getenv('VIRTUAL_ENV')
    if venv:
        return f'{venv}/bin/websphinx'
    d = split(Path(__file__).absolute())[0]
    return d + '/websphinx.py'


def init_browser_ext():
    cmd = get_executable()
    print(cmd)
    # init ff
    if exists(f'{Path.home()}/.mozilla/'):
        nm_dir = f'{Path.home()}/.mozilla/native-messaging-hosts/'
        if not exists(nm_dir):
            mkdir(nm_dir)
        with open(nm_dir+'websphinx.json', 'w') as outfile:
            ff_nm = FF_NM_TPL.format(cmd=cmd)
            outfile.write(ff_nm)
    # init chrome
    if exists(f'{Path.home()}/.config/chromium'):
        nm_dir = f'{Path.home()}/.config/chromium/NativeMessagingHosts/'
        if not exists(nm_dir):
            mkdir(nm_dir)
        with open(nm_dir+'websphinx.json', 'w') as outfile:
            chr_nm = CHR_NM_TPL.format(cmd=cmd)
            outfile.write(chr_nm)



import unittest
from os import listdir
from shutil import rmtree
from unittest.mock import Mock
from io import BytesIO
import sys, pysodium

from pwdsphinx import sphinx, bin2pass

# to get coverage, run
# PYTHONPATH=.. coverage run ../tests/rules.py
# coverage report -m
# to just run the tests do
# python3 -m unittest discover --start-directory ../tests

def equ(classes, syms, size, check, xor):
    unpacked = sphinx.unpack_rule(sphinx.pack_rule(classes, syms, size, check, xor))
    assert set(classes) == unpacked[0]
    assert list(syms) == unpacked[1]
    assert size == unpacked[2]
    if sphinx.validate_password:
        assert check == unpacked[3]
    else:
        assert 0 == unpacked[3]
    assert xor == unpacked[4]

from itertools import chain, combinations
def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

class TestRules(unittest.TestCase):
    def test_rules(self):
       for cls in powerset('uld'):
           equ(''.join(cls), bin2pass.symbols, 64, 31, b'\x00'*64)
           if cls!=tuple():
               equ(''.join(cls), '', 64, 31, b'\x00'*64)

       equ('uld', bin2pass.symbols[:16], 64, 31, b'\x00'*64)
       equ('uld', bin2pass.symbols[16:], 64, 31, b'\x00'*64)

       equ('uld', bin2pass.symbols, 64, 31, b'\xff'*64)
       equ('uld', bin2pass.symbols, 64, 31, b'\xaa'*64)

       for i in range(128):
           equ('uld', bin2pass.symbols, i, 31, b'\xaa'*64)
       for i in range(32):
           equ('uld', bin2pass.symbols, 64, i, b'\xaa'*64)

if __name__ == '__main__':
    unittest.main()
